@@ -1,3 +1,4 @@
+#include <inttypes.h>
 #include <sys/wait.h>
 #include <sys/types.h>
 #include <unistd.h>
@@ -27,16 +28,17 @@
 #include "rpc-pie-priv.h"
 #include "infect-util.h"
 
-#define __sys(foo)	foo
-#define __sys_err(ret)	(-errno)
+#define __sys(foo)     foo
+#define __sys_err(ret) (-errno)
 
 #include "common/scm.h"
 #include "common/scm-code.c"
 
-#define UNIX_PATH_MAX (sizeof(struct sockaddr_un) - \
-			(size_t)((struct sockaddr_un *) 0)->sun_path)
+#ifndef UNIX_PATH_MAX
+#define UNIX_PATH_MAX (sizeof(struct sockaddr_un) - (size_t)((struct sockaddr_un *)0)->sun_path)
+#endif
 
-#define PARASITE_STACK_SIZE	(16 << 10)
+#define PARASITE_STACK_SIZE (16 << 10)
 
 #ifndef SECCOMP_MODE_DISABLED
 #define SECCOMP_MODE_DISABLED 0
@@ -90,6 +92,12 @@ static int parse_pid_status(int pid, struct seize_task_status *ss, void *data)
 
 			continue;
 		}
+		if (!strncmp(aux, "SigBlk:", 7)) {
+			if (sscanf(aux + 7, "%llx", &ss->sigblk) != 1)
+				goto err_parse;
+
+			continue;
+		}
 	}
 
 	fclose(f);
@@ -103,7 +111,7 @@ err_parse:
 int compel_stop_task(int pid)
 {
 	int ret;
-	struct seize_task_status ss;
+	struct seize_task_status ss = {};
 
 	ret = compel_interrupt_task(pid);
 	if (ret == 0)
@@ -137,7 +145,7 @@ int compel_interrupt_task(int pid)
 	 */
 	ret = ptrace(PTRACE_INTERRUPT, pid, NULL, NULL);
 	if (ret < 0) {
-		pr_warn("SEIZE %d: can't interrupt task: %s", pid, strerror(errno));
+		pr_warn("SEIZE %d: can't interrupt task: %s\n", pid, strerror(errno));
 		if (ptrace(PTRACE_DETACH, pid, NULL, NULL))
 			pr_perror("Unable to detach from %d", pid);
 	}
@@ -184,6 +192,29 @@ static int skip_sigstop(int pid, int nr_signals)
 	return 0;
 }
 
+#define SIG_MASK(sig) (1ULL << ((sig)-1))
+
+#define SIG_IN_MASK(sig, mask) ((sig) > 0 && (sig) <= SIGMAX && (SIG_MASK(sig) & (mask)))
+
+#define SUPPORTED_STOP_MASK ((1ULL << (SIGSTOP - 1)) | (1ULL << (SIGTSTP - 1)))
+
+static inline int sig_stop(int sig)
+{
+	return SIG_IN_MASK(sig, SUPPORTED_STOP_MASK);
+}
+
+int compel_parse_stop_signo(int pid)
+{
+	siginfo_t si;
+
+	if (ptrace(PTRACE_GETSIGINFO, pid, NULL, &si) < 0) {
+		pr_perror("SEIZE %d: can't parse stopped siginfo", pid);
+		return -1;
+	}
+
+	return si.si_signo;
+}
+
 /*
  * This routine seizes task putting it into a special
  * state where we can manipulate the task via ptrace
@@ -191,13 +222,12 @@ static int skip_sigstop(int pid, int nr_signals)
  * of it so the task would not know if it was saddled
  * up with someone else.
  */
-int compel_wait_task(int pid, int ppid,
-		int (*get_status)(int pid, struct seize_task_status *, void *),
-		void (*free_status)(int pid, struct seize_task_status *, void *),
-		struct seize_task_status *ss, void *data)
+int compel_wait_task(int pid, int ppid, int (*get_status)(int pid, struct seize_task_status *, void *),
+		     void (*free_status)(int pid, struct seize_task_status *, void *), struct seize_task_status *ss,
+		     void *data)
 {
 	siginfo_t si;
-	int status, nr_sigstop;
+	int status, nr_stopsig;
 	int ret = 0, ret2, wait_errno = 0;
 
 	/*
@@ -230,8 +260,8 @@ try_again:
 			if (pid == getpid())
 				pr_err("The criu itself is within dumped tree.\n");
 			else
-				pr_err("Unseizable non-zombie %d found, state %c, err %d/%d\n",
-						pid, ss->state, ret, wait_errno);
+				pr_err("Unseizable non-zombie %d found, state %c, err %d/%d\n", pid, ss->state, ret,
+				       wait_errno);
 			return -1;
 		}
 
@@ -242,8 +272,7 @@ try_again:
 	}
 
 	if ((ppid != -1) && (ss->ppid != ppid)) {
-		pr_err("Task pid reused while suspending (%d: %d -> %d)\n",
-				pid, ppid, ss->ppid);
+		pr_err("Task pid reused while suspending (%d: %d -> %d)\n", pid, ppid, ss->ppid);
 		goto err;
 	}
 
@@ -265,32 +294,63 @@ try_again:
 		 * handle one and repeat.
 		 */
 
-		if (ptrace(PTRACE_CONT, pid, NULL,
-					(void *)(unsigned long)si.si_signo)) {
+		if (ptrace(PTRACE_CONT, pid, NULL, (void *)(unsigned long)si.si_signo)) {
 			pr_perror("Can't continue signal handling, aborting");
 			goto err;
 		}
 
-		ret = 0;
 		if (free_status)
 			free_status(pid, ss, data);
 		goto try_again;
 	}
 
+	if (ptrace(PTRACE_SETOPTIONS, pid, NULL, PTRACE_O_TRACESYSGOOD)) {
+		pr_perror("Unable to set PTRACE_O_TRACESYSGOOD for %d", pid);
+		return -1;
+	}
+
 	if (ss->seccomp_mode != SECCOMP_MODE_DISABLED && ptrace_suspend_seccomp(pid) < 0)
 		goto err;
 
-	nr_sigstop = 0;
-	if (ss->sigpnd & (1 << (SIGSTOP - 1)))
-		nr_sigstop++;
-	if (ss->shdpnd & (1 << (SIGSTOP - 1)))
-		nr_sigstop++;
-	if (si.si_signo == SIGSTOP)
-		nr_sigstop++;
+	/*
+	 * FIXME(issues/1429): parasite code contains instructions that trigger
+	 * SIGTRAP to stop at certain points. In such cases, the kernel sends a
+	 * force SIGTRAP that can't be ignored and if it is blocked, the kernel
+	 * resets its signal handler to a default one and unblocks it. It means
+	 * that if we want to save the origin signal handler, we need to run a
+	 * parasite code with the unblocked SIGTRAP.
+	 */
+	if ((ss->sigpnd | ss->shdpnd) & (1 << (SIGTRAP - 1))) {
+		pr_err("Can't dump the %d thread with a pending SIGTRAP.\n", pid);
+		goto err;
+	}
 
-	if (nr_sigstop) {
-		if (skip_sigstop(pid, nr_sigstop))
-			goto err_stop;
+	nr_stopsig = 0;
+	if (SIG_IN_MASK(SIGSTOP, ss->sigpnd))
+		nr_stopsig++;
+	if (SIG_IN_MASK(SIGSTOP, ss->shdpnd))
+		nr_stopsig++;
+
+	if (SIG_IN_MASK(SIGTSTP, ss->sigpnd) && !SIG_IN_MASK(SIGTSTP, ss->sigblk))
+		nr_stopsig++;
+	if (SIG_IN_MASK(SIGTSTP, ss->shdpnd) && !SIG_IN_MASK(SIGTSTP, ss->sigblk))
+		nr_stopsig++;
+
+	if (sig_stop(si.si_signo))
+		nr_stopsig++;
+
+	if (nr_stopsig) {
+		if (skip_sigstop(pid, nr_stopsig)) {
+			/*
+			 * Make sure that the task is stopped by a supported stop signal and
+			 * send it again to restore task state before criu intervention.
+			 */
+			if (sig_stop(si.si_signo))
+				kill(pid, si.si_signo);
+			else
+				kill(pid, SIGSTOP);
+			goto err;
+		}
 
 		return COMPEL_TASK_STOPPED;
 	}
@@ -302,8 +362,6 @@ try_again:
 		goto err;
 	}
 
-err_stop:
-	kill(pid, SIGSTOP);
 err:
 	if (ptrace(PTRACE_DETACH, pid, NULL, NULL))
 		pr_perror("Unable to detach from %d", pid);
@@ -312,6 +370,13 @@ err:
 
 int compel_resume_task(pid_t pid, int orig_st, int st)
 {
+	return compel_resume_task_sig(pid, orig_st, st, SIGSTOP);
+}
+
+int compel_resume_task_sig(pid_t pid, int orig_st, int st, int stop_signo)
+{
+	int ret = 0;
+
 	pr_debug("\tUnseizing %d into %d\n", pid, st);
 
 	if (st == COMPEL_TASK_DEAD) {
@@ -332,17 +397,29 @@ int compel_resume_task(pid_t pid, int orig_st, int st)
 		 * task with STOP in queue that would get lost after
 		 * detach, so stop it again.
 		 */
-		if (orig_st == COMPEL_TASK_STOPPED)
-			kill(pid, SIGSTOP);
-	} else
+		if (orig_st == COMPEL_TASK_STOPPED) {
+			/*
+			 * Check that stop_signo contain supported stop signal.
+			 * If it isn't, then send SIGSTOP. It makes sense in the case
+			 * when we get COMPEL_TASK_STOPPED from old image,
+			 * where stop_signo was not yet supported.
+			 */
+			if (sig_stop(stop_signo))
+				kill(pid, stop_signo);
+			else
+				kill(pid, SIGSTOP);
+		}
+	} else {
 		pr_err("Unknown final state %d\n", st);
+		ret = -1;
+	}
 
 	if (ptrace(PTRACE_DETACH, pid, NULL, NULL)) {
 		pr_perror("Unable to detach from %d", pid);
 		return -1;
 	}
 
-	return 0;
+	return ret;
 }
 
 static int gen_parasite_saddr(struct sockaddr_un *saddr, int key)
@@ -350,8 +427,7 @@ static int gen_parasite_saddr(struct sockaddr_un *saddr, int key)
 	int sun_len;
 
 	saddr->sun_family = AF_UNIX;
-	snprintf(saddr->sun_path, UNIX_PATH_MAX,
-			"X/crtools-pr-%d", key);
+	snprintf(saddr->sun_path, UNIX_PATH_MAX, "X/crtools-pr-%d-%" PRIx64, key, compel_run_id);
 
 	sun_len = SUN_LEN(saddr);
 	*saddr->sun_path = '\0';
@@ -359,8 +435,7 @@ static int gen_parasite_saddr(struct sockaddr_un *saddr, int key)
 	return sun_len;
 }
 
-static int prepare_tsock(struct parasite_ctl *ctl, pid_t pid,
-		struct parasite_init_args *args)
+static int prepare_tsock(struct parasite_ctl *ctl, pid_t pid, struct parasite_init_args *args)
 {
 	int ssock = -1;
 	socklen_t sk_len;
@@ -377,7 +452,7 @@ static int prepare_tsock(struct parasite_ctl *ctl, pid_t pid,
 		goto err;
 	}
 
-	if (getsockname(ssock, (struct sockaddr *) &addr, &sk_len) < 0) {
+	if (getsockname(ssock, (struct sockaddr *)&addr, &sk_len) < 0) {
 		pr_perror("Unable to get name for a socket");
 		return -1;
 	}
@@ -412,8 +487,8 @@ err:
 static int setup_child_handler(struct parasite_ctl *ctl)
 {
 	struct sigaction sa = {
-		.sa_sigaction	= ctl->ictx.child_handler,
-		.sa_flags	= SA_SIGINFO | SA_RESTART,
+		.sa_sigaction = ctl->ictx.child_handler,
+		.sa_flags = SA_SIGINFO | SA_RESTART,
 	};
 
 	sigemptyset(&sa.sa_mask);
@@ -436,12 +511,17 @@ static int restore_child_handler(struct parasite_ctl *ctl)
 	return 0;
 }
 
-static int parasite_run(pid_t pid, int cmd, unsigned long ip, void *stack,
-		user_regs_struct_t *regs, struct thread_ctx *octx)
+static int parasite_run(pid_t pid, int cmd, unsigned long ip, void *stack, user_regs_struct_t *regs,
+			struct thread_ctx *octx)
 {
 	k_rtsigset_t block;
 
 	ksigfillset(&block);
+	/*
+	 * FIXME(issues/1429): SIGTRAP can't be blocked, otherwise its handler
+	 * will be reset to the default one.
+	 */
+	ksigdelset(&block, SIGTRAP);
 	if (ptrace(PTRACE_SETSIGMASK, pid, sizeof(k_rtsigset_t), &block)) {
 		pr_perror("Can't block signals for %d", pid);
 		goto err_sig;
@@ -470,7 +550,7 @@ err_sig:
 	return -1;
 }
 
-static int restore_thread_ctx(int pid, struct thread_ctx *ctx)
+static int restore_thread_ctx(int pid, struct thread_ctx *ctx, bool restore_ext_regs)
 {
 	int ret = 0;
 
@@ -478,6 +558,10 @@ static int restore_thread_ctx(int pid, struct thread_ctx *ctx)
 		pr_perror("Can't restore registers (pid: %d)", pid);
 		ret = -1;
 	}
+
+	if (restore_ext_regs && compel_set_task_ext_regs(pid, &ctx->ext_regs))
+		ret = -1;
+
 	if (ptrace(PTRACE_SETSIGMASK, pid, sizeof(k_rtsigset_t), &ctx->sigmask)) {
 		pr_perror("Can't block signals");
 		ret = -1;
@@ -486,11 +570,9 @@ static int restore_thread_ctx(int pid, struct thread_ctx *ctx)
 	return ret;
 }
 
-
 /* we run at @regs->ip */
-static int parasite_trap(struct parasite_ctl *ctl, pid_t pid,
-				user_regs_struct_t *regs,
-				struct thread_ctx *octx)
+static int parasite_trap(struct parasite_ctl *ctl, pid_t pid, user_regs_struct_t *regs, struct thread_ctx *octx,
+			 bool may_use_extended_regs)
 {
 	siginfo_t siginfo;
 	int status;
@@ -507,7 +589,7 @@ static int parasite_trap(struct parasite_ctl *ctl, pid_t pid,
 	}
 
 	if (!WIFSTOPPED(status)) {
-		pr_err("Task is still running (pid: %d)\n", pid);
+		pr_err("Task is still running (pid: %d, status: 0x%x)\n", pid, status);
 		goto err;
 	}
 
@@ -518,12 +600,11 @@ static int parasite_trap(struct parasite_ctl *ctl, pid_t pid,
 
 	if (ptrace_get_regs(pid, regs)) {
 		pr_perror("Can't obtain registers (pid: %d)", pid);
-			goto err;
+		goto err;
 	}
 
 	if (WSTOPSIG(status) != SIGTRAP || siginfo.si_code != ARCH_SI_TRAP) {
-		pr_debug("** delivering signal %d si_code=%d\n",
-			 siginfo.si_signo, siginfo.si_code);
+		pr_debug("** delivering signal %d si_code=%d\n", siginfo.si_signo, siginfo.si_code);
 
 		pr_err("Unexpected %d task interruption, aborting\n", pid);
 		goto err;
@@ -535,15 +616,13 @@ static int parasite_trap(struct parasite_ctl *ctl, pid_t pid,
 	 */
 	ret = 0;
 err:
-	if (restore_thread_ctx(pid, octx))
+	if (restore_thread_ctx(pid, octx, may_use_extended_regs))
 		ret = -1;
 
 	return ret;
 }
 
-
-int compel_execute_syscall(struct parasite_ctl *ctl,
-		user_regs_struct_t *regs, const char *code_syscall)
+int compel_execute_syscall(struct parasite_ctl *ctl, user_regs_struct_t *regs, const char *code_syscall)
 {
 	pid_t pid = ctl->rpid;
 	int err;
@@ -554,18 +633,16 @@ int compel_execute_syscall(struct parasite_ctl *ctl,
 	 * we will need it to restore original program content.
 	 */
 	memcpy(code_orig, code_syscall, sizeof(code_orig));
-	if (ptrace_swap_area(pid, (void *)ctl->ictx.syscall_ip,
-			     (void *)code_orig, sizeof(code_orig))) {
+	if (ptrace_swap_area(pid, (void *)ctl->ictx.syscall_ip, (void *)code_orig, sizeof(code_orig))) {
 		pr_err("Can't inject syscall blob (pid: %d)\n", pid);
 		return -1;
 	}
 
 	err = parasite_run(pid, PTRACE_CONT, ctl->ictx.syscall_ip, 0, regs, &ctl->orig);
 	if (!err)
-		err = parasite_trap(ctl, pid, regs, &ctl->orig);
+		err = parasite_trap(ctl, pid, regs, &ctl->orig, false);
 
-	if (ptrace_poke_area(pid, (void *)code_orig,
-			     (void *)ctl->ictx.syscall_ip, sizeof(code_orig))) {
+	if (ptrace_poke_area(pid, (void *)code_orig, (void *)ctl->ictx.syscall_ip, sizeof(code_orig))) {
 		pr_err("Can't restore syscall blob (pid: %d)\n", ctl->rpid);
 		err = -1;
 	}
@@ -580,7 +657,7 @@ int compel_run_at(struct parasite_ctl *ctl, unsigned long ip, user_regs_struct_t
 
 	ret = parasite_run(ctl->rpid, PTRACE_CONT, ip, 0, &regs, &ctl->orig);
 	if (!ret)
-		ret = parasite_trap(ctl, ctl->rpid, ret_regs ? ret_regs : &regs, &ctl->orig);
+		ret = parasite_trap(ctl, ctl->rpid, ret_regs ? ret_regs : &regs, &ctl->orig, false);
 	return ret;
 }
 
@@ -605,14 +682,17 @@ static int parasite_init_daemon(struct parasite_ctl *ctl)
 	struct parasite_init_args *args;
 	pid_t pid = ctl->rpid;
 	user_regs_struct_t regs;
-	struct ctl_msg m = { };
+	struct ctl_msg m = {};
 
-	*ctl->addr_cmd = PARASITE_CMD_INIT_DAEMON;
+	*ctl->cmd = PARASITE_CMD_INIT_DAEMON;
 
 	args = compel_parasite_args(ctl, struct parasite_init_args);
 
 	args->sigframe = (uintptr_t)ctl->rsigframe;
 	args->log_level = compel_log_get_loglevel();
+#ifdef ARCH_HAS_LONG_PAGES
+	args->page_size = PAGE_SIZE;
+#endif
 
 	futex_set(&args->daemon_connected, 0);
 
@@ -643,12 +723,11 @@ static int parasite_init_daemon(struct parasite_ctl *ctl)
 	pr_info("Wait for parasite being daemonized...\n");
 
 	if (parasite_wait_ack(ctl->tsock, PARASITE_CMD_INIT_DAEMON, &m)) {
-		pr_err("Can't switch parasite %d to daemon mode %d\n",
-		       pid, m.err);
+		pr_err("Can't switch parasite %d to daemon mode %d\n", pid, m.err);
 		goto err;
 	}
 
-	ctl->sigreturn_addr = (void*)(uintptr_t)args->sigreturn_addr;
+	ctl->sigreturn_addr = (void *)(uintptr_t)args->sigreturn_addr;
 	ctl->daemonized = true;
 	pr_info("Parasite %d has been switched to daemon mode\n", pid);
 	return 0;
@@ -663,16 +742,20 @@ static int parasite_start_daemon(struct parasite_ctl *ctl)
 
 	/*
 	 * Get task registers before going daemon, since the
-	 * compel_get_task_regs needs to call ptrace on _stopped_ task,
+	 * compel_get_task_regs() needs to call ptrace on _stopped_ task,
 	 * while in daemon it is not such.
 	 */
 
-	if (get_task_regs(pid, &ctl->orig.regs, ictx->save_regs,
-			ictx->regs_arg, ictx->flags)) {
+	if (compel_get_task_regs(pid, &ctl->orig.regs, NULL, ictx->save_regs, ictx->regs_arg, ictx->flags)) {
 		pr_err("Can't obtain regs for thread %d\n", pid);
 		return -1;
 	}
 
+	if (__compel_arch_fetch_thread_area(pid, &ctl->orig)) {
+		pr_err("Can't get thread area of %d\n", pid);
+		return -1;
+	}
+
 	if (ictx->make_sigframe(ictx->regs_arg, ctl->sigframe, ctl->rsigframe, &ctl->orig.sigmask))
 		return -1;
 
@@ -682,13 +765,11 @@ static int parasite_start_daemon(struct parasite_ctl *ctl)
 	return 0;
 }
 
-static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size)
+static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size, int remote_prot)
 {
 	int fd;
 
-	ctl->remote_map = remote_mmap(ctl, NULL, size,
-				      PROT_READ | PROT_WRITE | PROT_EXEC,
-				      MAP_ANONYMOUS | MAP_SHARED, -1, 0);
+	ctl->remote_map = remote_mmap(ctl, NULL, size, remote_prot, MAP_ANONYMOUS | MAP_SHARED, -1, 0);
 	if (!ctl->remote_map) {
 		pr_err("Can't allocate memory for parasite blob (pid: %d)\n", ctl->rpid);
 		return -1;
@@ -696,13 +777,12 @@ static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size)
 
 	ctl->map_length = round_up(size, page_size());
 
-	fd = ctl->ictx.open_proc(ctl->rpid, O_RDWR, "map_files/%lx-%lx",
-		 (long)ctl->remote_map, (long)ctl->remote_map + ctl->map_length);
+	fd = ctl->ictx.open_proc(ctl->rpid, O_RDWR, "map_files/%lx-%lx", (long)ctl->remote_map,
+				 (long)ctl->remote_map + ctl->map_length);
 	if (fd < 0)
 		return -1;
 
-	ctl->local_map = mmap(NULL, size, PROT_READ | PROT_WRITE,
-			      MAP_SHARED | MAP_FILE, fd, 0);
+	ctl->local_map = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FILE, fd, 0);
 	close(fd);
 
 	if (ctl->local_map == MAP_FAILED) {
@@ -714,14 +794,25 @@ static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size)
 	return 0;
 }
 
-static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
+static void parasite_memfd_close(struct parasite_ctl *ctl, int fd)
+{
+	bool compat = !compel_mode_native(ctl);
+	long ret;
+	int err;
+
+	err = compel_syscall(ctl, __NR(close, compat), &ret, fd, 0, 0, 0, 0, 0);
+	if (err || ret)
+		pr_err("Can't close memfd\n");
+}
+
+static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size, int remote_prot)
 {
 	void *where = (void *)ctl->ictx.syscall_ip + BUILTIN_SYSCALL_SIZE;
+	bool compat_task = !compel_mode_native(ctl);
 	uint8_t orig_code[MEMFD_FNAME_SZ] = MEMFD_FNAME;
 	pid_t pid = ctl->rpid;
 	long sret = -ENOSYS;
 	int ret, fd, lfd;
-	bool __maybe_unused compat_task = !compel_mode_native(ctl);
 
 	if (ctl->ictx.flags & INFECT_NO_MEMFD)
 		return 1;
@@ -733,14 +824,12 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 		return -1;
 	}
 
-	ret = compel_syscall(ctl, __NR(memfd_create, compat_task), &sret,
-			     (unsigned long)where, 0, 0, 0, 0, 0);
+	ret = compel_syscall(ctl, __NR(memfd_create, compat_task), &sret, (unsigned long)where, 0, 0, 0, 0, 0);
 
 	if (ptrace_poke_area(pid, orig_code, where, sizeof(orig_code))) {
-		fd = (int)(long)sret;
+		fd = (int)sret;
 		if (fd >= 0)
-			compel_syscall(ctl, __NR(close, compat_task), &sret,
-					fd, 0, 0, 0, 0, 0);
+			parasite_memfd_close(ctl, fd);
 		pr_err("Can't restore memfd args (pid: %d)\n", pid);
 		return -1;
 	}
@@ -748,7 +837,7 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 	if (ret < 0)
 		return ret;
 
-	fd = (int)(long)sret;
+	fd = (int)sret;
 	if (fd == -ENOSYS)
 		return 1;
 	if (fd < 0) {
@@ -767,23 +856,20 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 		goto err_cure;
 	}
 
-	ctl->remote_map = remote_mmap(ctl, NULL, size,
-				      PROT_READ | PROT_WRITE | PROT_EXEC,
-				      MAP_FILE | MAP_SHARED, fd, 0);
+	ctl->remote_map = remote_mmap(ctl, NULL, size, remote_prot, MAP_FILE | MAP_SHARED, fd, 0);
 	if (!ctl->remote_map) {
 		pr_err("Can't rmap memfd for parasite blob\n");
 		goto err_curef;
 	}
 
-	ctl->local_map = mmap(NULL, size, PROT_READ | PROT_WRITE,
-			      MAP_SHARED | MAP_FILE, lfd, 0);
+	ctl->local_map = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FILE, lfd, 0);
 	if (ctl->local_map == MAP_FAILED) {
 		ctl->local_map = NULL;
 		pr_perror("Can't lmap memfd for parasite blob");
 		goto err_curef;
 	}
 
-	compel_syscall(ctl, __NR(close, compat_task), &sret, fd, 0, 0, 0, 0, 0);
+	parasite_memfd_close(ctl, fd);
 	close(lfd);
 
 	pr_info("Set up parasite blob using memfd\n");
@@ -792,25 +878,39 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 err_curef:
 	close(lfd);
 err_cure:
-	compel_syscall(ctl, __NR(close, compat_task), &sret, fd, 0, 0, 0, 0, 0);
+	parasite_memfd_close(ctl, fd);
 	return -1;
 }
 
-void compel_relocs_apply(void *mem, void *vbase, size_t size, compel_reloc_t *elf_relocs, size_t nr_relocs)
+void compel_relocs_apply(void *mem, void *vbase, struct parasite_blob_desc *pbd)
 {
-	size_t i, j;
+	compel_reloc_t *elf_relocs = pbd->hdr.relocs;
+	size_t nr_relocs = pbd->hdr.nr_relocs;
 
+	size_t i, j;
+	void **got = mem + pbd->hdr.got_off;
+
+	/*
+	 * parasite_service() reads the value of __export_parasite_service_args_ptr.
+	 * The reason it is set here is that semantically, we are doing a symbol
+	 * resolution on parasite_service_args, and it turns out to be relocatable.
+	 */
+	*(void **)(mem + pbd->hdr.args_ptr_off) = vbase + pbd->hdr.args_off;
+
+#ifdef CONFIG_MIPS
+	compel_relocs_apply_mips(mem, vbase, pbd);
+#else
 	for (i = 0, j = 0; i < nr_relocs; i++) {
 		if (elf_relocs[i].type & COMPEL_TYPE_LONG) {
 			long *where = mem + elf_relocs[i].offset;
-			long *p = mem + size;
 
 			if (elf_relocs[i].type & COMPEL_TYPE_GOTPCREL) {
 				int *value = (int *)where;
 				int rel;
 
-				p[j] = (long)vbase + elf_relocs[i].value;
-				rel = (unsigned)((void *)&p[j] - (void *)mem) - elf_relocs[i].offset + elf_relocs[i].addend;
+				got[j] = vbase + elf_relocs[i].value;
+				rel = (unsigned)((void *)&got[j] - (void *)mem) - elf_relocs[i].offset +
+				      elf_relocs[i].addend;
 
 				*value = rel;
 				j++;
@@ -822,26 +922,52 @@ void compel_relocs_apply(void *mem, void *vbase, size_t size, compel_reloc_t *el
 		} else
 			BUG();
 	}
+#endif
 }
 
-static int compel_map_exchange(struct parasite_ctl *ctl, unsigned long size)
+long remote_mprotect(struct parasite_ctl *ctl, void *addr, size_t len, int prot)
 {
-	int ret;
+	long ret;
+	int err;
+	bool compat_task = !user_regs_native(&ctl->orig.regs);
 
-	ret = parasite_memfd_exchange(ctl, size);
-	if (ret == 1) {
-		pr_info("MemFD parasite doesn't work, goto legacy mmap\n");
-		ret = parasite_mmap_exchange(ctl, size);
+	err = compel_syscall(ctl, __NR(mprotect, compat_task), &ret, (unsigned long)addr, len, prot, 0, 0, 0);
+	if (err < 0) {
+		pr_err("compel_syscall for mprotect failed\n");
+		return -1;
 	}
 	return ret;
 }
 
-static inline unsigned long total_pie_size(size_t blob_size)
+static int compel_map_exchange(struct parasite_ctl *ctl, unsigned long size)
 {
-	return round_up(blob_size, page_size());
+	int ret, remote_prot;
+
+	if (ctl->pblob.hdr.data_off)
+		remote_prot = PROT_READ | PROT_EXEC;
+	else
+		remote_prot = PROT_READ | PROT_WRITE | PROT_EXEC;
+
+	ret = parasite_memfd_exchange(ctl, size, remote_prot);
+	if (ret == 1) {
+		pr_info("MemFD parasite doesn't work, goto legacy mmap\n");
+		ret = parasite_mmap_exchange(ctl, size, remote_prot);
+		if (ret)
+			return ret;
+	}
+
+	if (!ctl->pblob.hdr.data_off)
+		return 0;
+
+	ret = remote_mprotect(ctl, ctl->remote_map + ctl->pblob.hdr.data_off, size - ctl->pblob.hdr.data_off,
+			      PROT_READ | PROT_WRITE);
+	if (ret)
+		pr_err("remote_mprotect failed\n");
+
+	return ret;
 }
 
-int compel_infect(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned long args_size)
+int compel_infect_no_daemon(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned long args_size)
 {
 	int ret;
 	unsigned long p, map_exchange_size, parasite_size = 0;
@@ -862,11 +988,39 @@ int compel_infect(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned l
 	 * without using ptrace at all.
 	 */
 
-	parasite_size = total_pie_size(ctl->pblob.hdr.bsize);
+	/*
+	 * The parasite memory layout is the following:
+	 * Low address start first.
+	 * The number in parenthesis denotes the size of the section.
+	 * The arrow on the right shows the different variables that
+	 * corresponds to a given offset.
+	 * +------------------------------------------------------+ <--- 0
+	 * |   Parasite blob (sizeof(parasite_blob))              |
+	 * +------------------------------------------------------+ <--- hdr.bsize
+	 *                         align 8
+	 * +------------------------------------------------------+ <--- hdr.got_off
+	 * |   GOT Table (nr_gotpcrel * sizeof(long))             |
+	 * +------------------------------------------------------+ <--- hdr.args_off
+	 * |   Args area (args_size)                              |
+	 * +------------------------------------------------------+
+	 *                         align 64
+	 * +------------------------------------------------------+ <--- ctl->rsigframe
+	 * |   sigframe (RESTORE_STACK_SIGFRAME)                  |      ctl->sigframe
+	 * +------------------------------------------------------+
+	 * |   main stack (PARASITE_STACK_SIZE)                   |
+	 * +------------------------------------------------------+ <--- ctl->rstack
+	 * |   compel_run_in_thread stack (PARASITE_STACK_SIZE)   |
+	 * +------------------------------------------------------+ <--- ctl->r_thread_stack
+	 *                                                               map_exchange_size
+	 */
+	parasite_size = ctl->pblob.hdr.args_off;
 
-	ctl->args_size = round_up(args_size, PAGE_SIZE);
+	ctl->args_size = args_size;
 	parasite_size += ctl->args_size;
 
+	/* RESTORE_STACK_SIGFRAME needs a 64 bytes alignment */
+	parasite_size = round_up(parasite_size, 64);
+
 	map_exchange_size = parasite_size;
 	map_exchange_size += RESTORE_STACK_SIGFRAME + PARASITE_STACK_SIZE;
 	if (nr_threads > 1)
@@ -879,23 +1033,41 @@ int compel_infect(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned l
 	pr_info("Putting parasite blob into %p->%p\n", ctl->local_map, ctl->remote_map);
 
 	ctl->parasite_ip = (unsigned long)(ctl->remote_map + ctl->pblob.hdr.parasite_ip_off);
-	ctl->addr_cmd = ctl->local_map + ctl->pblob.hdr.addr_cmd_off;
-	ctl->addr_args = ctl->local_map + ctl->pblob.hdr.addr_arg_off;
+	ctl->cmd = ctl->local_map + ctl->pblob.hdr.cmd_off;
+	ctl->args = ctl->local_map + ctl->pblob.hdr.args_off;
+
+	/*
+	 * args must be 4 bytes aligned as we use futexes() on them. It is
+	 * already the case, as args follows the GOT table, which is 8 bytes
+	 * aligned.
+	 */
+	if ((unsigned long)ctl->args & (4 - 1)) {
+		pr_err("BUG: args are not 4 bytes aligned: %p\n", ctl->args);
+		goto err;
+	}
 
 	memcpy(ctl->local_map, ctl->pblob.hdr.mem, ctl->pblob.hdr.bsize);
-	if (ctl->pblob.hdr.nr_relocs)
-		compel_relocs_apply(ctl->local_map, ctl->remote_map, ctl->pblob.hdr.bsize,
-				    ctl->pblob.hdr.relocs, ctl->pblob.hdr.nr_relocs);
+	compel_relocs_apply(ctl->local_map, ctl->remote_map, &ctl->pblob);
 
 	p = parasite_size;
 
-	ctl->rsigframe	= ctl->remote_map + p;
-	ctl->sigframe	= ctl->local_map  + p;
+	ctl->rsigframe = ctl->remote_map + p;
+	ctl->sigframe = ctl->local_map + p;
 
 	p += RESTORE_STACK_SIGFRAME;
 	p += PARASITE_STACK_SIZE;
 	ctl->rstack = ctl->remote_map + p;
 
+	/*
+	 * x86-64 ABI requires a 16 bytes aligned stack.
+	 * It is already the case as RESTORE_STACK_SIGFRAME is a multiple of
+	 * 64, and PARASITE_STACK_SIZE is 0x4000.
+	 */
+	if ((unsigned long)ctl->rstack & (16 - 1)) {
+		pr_err("BUG: stack is not 16 bytes aligned: %p\n", ctl->rstack);
+		goto err;
+	}
+
 	if (nr_threads > 1) {
 		p += PARASITE_STACK_SIZE;
 		ctl->r_thread_stack = ctl->remote_map + p;
@@ -903,20 +1075,27 @@ int compel_infect(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned l
 
 	ret = arch_fetch_sas(ctl, ctl->rsigframe);
 	if (ret) {
-		pr_err("Can't fetch sigaltstack for task %d (ret %d)\n",
-		       ctl->rpid, ret);
+		pr_err("Can't fetch sigaltstack for task %d (ret %d)\n", ctl->rpid, ret);
 		goto err;
 	}
 
-	if (parasite_start_daemon(ctl))
-		goto err;
-
 	return 0;
 
 err:
 	return -1;
 }
 
+int compel_infect(struct parasite_ctl *ctl, unsigned long nr_threads, unsigned long args_size)
+{
+	if (compel_infect_no_daemon(ctl, nr_threads, args_size))
+		return -1;
+
+	if (parasite_start_daemon(ctl))
+		return -1;
+
+	return 0;
+}
+
 struct parasite_thread_ctl *compel_prepare_thread(struct parasite_ctl *ctl, int pid)
 {
 	struct parasite_thread_ctl *tctl;
@@ -1095,14 +1274,12 @@ static void handle_sigchld(int signal, siginfo_t *siginfo, void *data)
 	if (pid <= 0)
 		return;
 
-	pr_err("si_code=%d si_pid=%d si_status=%d\n",
-		siginfo->si_code, siginfo->si_pid, siginfo->si_status);
+	pr_err("si_code=%d si_pid=%d si_status=%d\n", siginfo->si_code, siginfo->si_pid, siginfo->si_status);
 
 	if (WIFEXITED(status))
 		pr_err("%d exited with %d unexpectedly\n", pid, WEXITSTATUS(status));
 	else if (WIFSIGNALED(status))
-		pr_err("%d was killed by %d unexpectedly: %s\n",
-			pid, WTERMSIG(status), strsignal(WTERMSIG(status)));
+		pr_err("%d was killed by %d unexpectedly: %s\n", pid, WTERMSIG(status), strsignal(WTERMSIG(status)));
 	else if (WIFSTOPPED(status))
 		pr_err("%d was stopped by %d unexpectedly\n", pid, WSTOPSIG(status));
 
@@ -1125,15 +1302,9 @@ static int save_regs_plain(void *to, user_regs_struct_t *r, user_fpregs_struct_t
 	return 0;
 }
 
-#ifndef RT_SIGFRAME_UC_SIGMASK
-#define RT_SIGFRAME_UC_SIGMASK(sigframe)				\
-	(k_rtsigset_t*)(void *)&RT_SIGFRAME_UC(sigframe)->uc_sigmask
-#endif
-
 static int make_sigframe_plain(void *from, struct rt_sigframe *f, struct rt_sigframe *rtf, k_rtsigset_t *b)
 {
 	struct plain_regs_struct *prs = from;
-	k_rtsigset_t *blk_sigset;
 
 	/*
 	 * Make sure it's zeroified.
@@ -1143,11 +1314,8 @@ static int make_sigframe_plain(void *from, struct rt_sigframe *f, struct rt_sigf
 	if (sigreturn_prep_regs_plain(f, &prs->regs, &prs->fpregs))
 		return -1;
 
-	blk_sigset = RT_SIGFRAME_UC_SIGMASK(f);
 	if (b)
-		memcpy(blk_sigset, b, sizeof(k_rtsigset_t));
-	else
-		memset(blk_sigset, 0, sizeof(k_rtsigset_t));
+		rt_sigframe_copy_sigset(f, b);
 
 	if (RT_SIGFRAME_HAS_FPU(f)) {
 		if (sigreturn_prep_fpu_frame_plain(f, rtf))
@@ -1202,9 +1370,8 @@ err:
 
 static bool task_in_parasite(struct parasite_ctl *ctl, user_regs_struct_t *regs)
 {
-	void *addr = (void *) REG_IP(*regs);
-	return addr >= ctl->remote_map &&
-		addr < ctl->remote_map + ctl->map_length;
+	void *addr = (void *)REG_IP(*regs);
+	return addr >= ctl->remote_map && addr < ctl->remote_map + ctl->map_length;
 }
 
 static int parasite_fini_seized(struct parasite_ctl *ctl)
@@ -1212,7 +1379,6 @@ static int parasite_fini_seized(struct parasite_ctl *ctl)
 	pid_t pid = ctl->rpid;
 	user_regs_struct_t regs;
 	int status, ret = 0;
-	enum trace_flags flag;
 
 	/* stop getting chld from parasite -- we're about to step-by-step it */
 	if (restore_child_handler(ctl))
@@ -1232,7 +1398,7 @@ static int parasite_fini_seized(struct parasite_ctl *ctl)
 
 	pr_debug("Daemon %d exited trapping\n", pid);
 	if (!WIFSTOPPED(status)) {
-		pr_err("Task is still running (pid: %d)\n", pid);
+		pr_err("Task is still running (pid: %d, status: 0x%x)\n", pid, status);
 		return -1;
 	}
 
@@ -1253,16 +1419,11 @@ static int parasite_fini_seized(struct parasite_ctl *ctl)
 		return -1;
 
 	/* Go to sigreturn as closer as we can */
-	ret = compel_stop_pie(pid, ctl->sigreturn_addr, &flag,
-			ctl->ictx.flags & INFECT_NO_BREAKPOINTS);
+	ret = compel_stop_pie(pid, ctl->sigreturn_addr, ctl->ictx.flags & INFECT_NO_BREAKPOINTS);
 	if (ret < 0)
 		return ret;
 
-	if (compel_stop_on_syscall(1, __NR(rt_sigreturn, 0),
-				__NR(rt_sigreturn, 1), flag))
-		return -1;
-
-	if (ptrace_flush_breakpoints(pid))
+	if (compel_stop_on_syscall(1, __NR(rt_sigreturn, 0), __NR(rt_sigreturn, 1)))
 		return -1;
 
 	/*
@@ -1274,6 +1435,11 @@ static int parasite_fini_seized(struct parasite_ctl *ctl)
 	return 0;
 }
 
+int compel_start_daemon(struct parasite_ctl *ctl)
+{
+	return parasite_start_daemon(ctl);
+}
+
 int compel_stop_daemon(struct parasite_ctl *ctl)
 {
 	if (ctl->daemonized) {
@@ -1298,6 +1464,7 @@ int compel_stop_daemon(struct parasite_ctl *ctl)
 int compel_cure_remote(struct parasite_ctl *ctl)
 {
 	long ret;
+	int err;
 
 	if (compel_stop_daemon(ctl))
 		return -1;
@@ -1305,12 +1472,13 @@ int compel_cure_remote(struct parasite_ctl *ctl)
 	if (!ctl->remote_map)
 		return 0;
 
-	compel_syscall(ctl, __NR(munmap, !compel_mode_native(ctl)), &ret,
-			(unsigned long)ctl->remote_map, ctl->map_length,
-			0, 0, 0, 0);
+	err = compel_syscall(ctl, __NR(munmap, !compel_mode_native(ctl)), &ret, (unsigned long)ctl->remote_map,
+			     ctl->map_length, 0, 0, 0, 0);
+	if (err)
+		return err;
+
 	if (ret) {
-		pr_err("munmap for remote map %p, %lu returned %lu\n",
-				ctl->remote_map, ctl->map_length, ret);
+		pr_err("munmap for remote map %p, %lu returned %lu\n", ctl->remote_map, ctl->map_length, ret);
 		return -1;
 	}
 
@@ -1345,10 +1513,10 @@ int compel_cure(struct parasite_ctl *ctl)
 
 void *compel_parasite_args_p(struct parasite_ctl *ctl)
 {
-	return ctl->addr_args;
+	return ctl->args;
 }
 
-void *compel_parasite_args_s(struct parasite_ctl *ctl, int args_size)
+void *compel_parasite_args_s(struct parasite_ctl *ctl, unsigned long args_size)
 {
 	BUG_ON(args_size > ctl->args_size);
 	return compel_parasite_args_p(ctl);
@@ -1363,11 +1531,11 @@ int compel_run_in_thread(struct parasite_thread_ctl *tctl, unsigned int cmd)
 	user_regs_struct_t regs = octx->regs;
 	int ret;
 
-	*ctl->addr_cmd = cmd;
+	*ctl->cmd = cmd;
 
 	ret = parasite_run(pid, PTRACE_CONT, ctl->parasite_ip, stack, &regs, octx);
 	if (ret == 0)
-		ret = parasite_trap(ctl, pid, &regs, octx);
+		ret = parasite_trap(ctl, pid, &regs, octx, true);
 	if (ret == 0)
 		ret = (int)REG_RES(regs);
 
@@ -1379,7 +1547,7 @@ int compel_run_in_thread(struct parasite_thread_ctl *tctl, unsigned int cmd)
 
 /*
  * compel_unmap() is used for unmapping parasite and restorer blobs.
- * A blob can contain code for unmapping itself, so the porcess is
+ * A blob can contain code for unmapping itself, so the process is
  * trapped on the exit from the munmap syscall.
  */
 int compel_unmap(struct parasite_ctl *ctl, unsigned long addr)
@@ -1392,21 +1560,24 @@ int compel_unmap(struct parasite_ctl *ctl, unsigned long addr)
 	if (ret)
 		goto err;
 
-	ret = compel_stop_on_syscall(1, __NR(munmap, 0),
-			__NR(munmap, 1), TRACE_ENTER);
+	ret = compel_stop_on_syscall(1, __NR(munmap, 0), __NR(munmap, 1));
 
-	if (restore_thread_ctx(pid, &ctl->orig))
+	/*
+	 * Don't touch extended registers here: they were restored
+	 * with rt_sigreturn from sigframe.
+	 */
+	if (restore_thread_ctx(pid, &ctl->orig, false))
 		ret = -1;
 err:
 	return ret;
 }
 
-int compel_stop_pie(pid_t pid, void *addr, enum trace_flags *tf, bool no_bp)
+int compel_stop_pie(pid_t pid, void *addr, bool no_bp)
 {
 	int ret;
 
 	if (no_bp) {
-		pr_debug("Force no-breakpoints restore\n");
+		pr_debug("Force no-breakpoints restore of %d\n", pid);
 		ret = 0;
 	} else
 		ret = ptrace_set_breakpoint(pid, addr);
@@ -1418,7 +1589,6 @@ int compel_stop_pie(pid_t pid, void *addr, enum trace_flags *tf, bool no_bp)
 		 * PIE will stop on a breakpoint, next
 		 * stop after that will be syscall enter.
 		 */
-		*tf = TRACE_EXIT;
 		return 0;
 	}
 
@@ -1431,39 +1601,34 @@ int compel_stop_pie(pid_t pid, void *addr, enum trace_flags *tf, bool no_bp)
 		pr_perror("Unable to restart the %d process", pid);
 		return -1;
 	}
-
-	*tf = TRACE_ENTER;
 	return 0;
 }
 
 static bool task_is_trapped(int status, pid_t pid)
 {
-	if (WIFSTOPPED(status) && WSTOPSIG(status) == SIGTRAP)
+	if (WIFSTOPPED(status) && (WSTOPSIG(status) & ~PTRACE_SYSCALL_TRAP) == SIGTRAP)
 		return true;
 
 	pr_err("Task %d is in unexpected state: %x\n", pid, status);
 	if (WIFEXITED(status))
 		pr_err("Task exited with %d\n", WEXITSTATUS(status));
 	if (WIFSIGNALED(status))
-		pr_err("Task signaled with %d: %s\n",
-			WTERMSIG(status), strsignal(WTERMSIG(status)));
+		pr_err("Task signaled with %d: %s\n", WTERMSIG(status), strsignal(WTERMSIG(status)));
 	if (WIFSTOPPED(status))
-		pr_err("Task stopped with %d: %s\n",
-			WSTOPSIG(status), strsignal(WSTOPSIG(status)));
+		pr_err("Task stopped with %d: %s\n", WSTOPSIG(status), strsignal(WSTOPSIG(status)));
 	if (WIFCONTINUED(status))
 		pr_err("Task continued\n");
 
 	return false;
 }
 
-static inline int is_required_syscall(user_regs_struct_t *regs, pid_t pid,
-		const int sys_nr, const int sys_nr_compat)
+static inline int is_required_syscall(user_regs_struct_t *regs, pid_t pid, const int sys_nr, const int sys_nr_compat)
 {
 	const char *mode = user_regs_native(regs) ? "native" : "compat";
 	int req_sysnr = user_regs_native(regs) ? sys_nr : sys_nr_compat;
 
-	pr_debug("%d (%s) is going to execute the syscall %lu, required is %d\n",
-		pid, mode, REG_SYSCALL_NR(*regs), req_sysnr);
+	pr_debug("%d (%s) is going to execute the syscall %lu, required is %d\n", pid, mode, REG_SYSCALL_NR(*regs),
+		 req_sysnr);
 
 	return (REG_SYSCALL_NR(*regs) == req_sysnr);
 }
@@ -1475,17 +1640,13 @@ static inline int is_required_syscall(user_regs_struct_t *regs, pid_t pid,
  * sys_nr - the required syscall number
  * sys_nr_compat - the required compatible syscall number
  */
-int compel_stop_on_syscall(int tasks,
-	const int sys_nr, const int sys_nr_compat,
-	enum trace_flags trace)
+int compel_stop_on_syscall(int tasks, const int sys_nr, const int sys_nr_compat)
 {
+	enum trace_flags trace = tasks > 1 ? TRACE_ALL : TRACE_ENTER;
 	user_regs_struct_t regs;
 	int status, ret;
 	pid_t pid;
 
-	if (tasks > 1)
-		trace = TRACE_ALL;
-
 	/* Stop all threads on the enter point in sys_rt_sigreturn */
 	while (tasks) {
 		pid = wait4(-1, &status, __WALL, NULL);
@@ -1499,6 +1660,18 @@ int compel_stop_on_syscall(int tasks,
 
 		pr_debug("%d was trapped\n", pid);
 
+		if ((WSTOPSIG(status) & PTRACE_SYSCALL_TRAP) == 0) {
+			/*
+			 * On some platforms such as ARM64, it is impossible to
+			 * pass through a breakpoint, so let's clear it right
+			 * after it has been triggered.
+			*/
+			if (ptrace_flush_breakpoints(pid)) {
+				pr_err("Unable to clear breakpoints\n");
+				return -1;
+			}
+			goto goon;
+		}
 		if (trace == TRACE_EXIT) {
 			trace = TRACE_ENTER;
 			pr_debug("`- Expecting exit\n");
@@ -1537,7 +1710,7 @@ int compel_stop_on_syscall(int tasks,
 			tasks--;
 			continue;
 		}
-goon:
+	goon:
 		ret = ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
 		if (ret) {
 			pr_perror("ptrace");
@@ -1568,9 +1741,9 @@ k_rtsigset_t *compel_task_sigmask(struct parasite_ctl *ctl)
 	return thread_ctx_sigmask(&ctl->orig);
 }
 
-int compel_get_thread_regs(struct parasite_thread_ctl *tctl, save_regs_t save, void * arg)
+int compel_get_thread_regs(struct parasite_thread_ctl *tctl, save_regs_t save, void *arg)
 {
-	return get_task_regs(tctl->tid, &tctl->th.regs, save, arg, tctl->ctl->ictx.flags);
+	return compel_get_task_regs(tctl->tid, &tctl->th.regs, &tctl->th.ext_regs, save, arg, tctl->ctl->ictx.flags);
 }
 
 struct infect_ctx *compel_infect_ctx(struct parasite_ctl *ctl)
@@ -1582,3 +1755,41 @@ struct parasite_blob_desc *compel_parasite_blob_desc(struct parasite_ctl *ctl)
 {
 	return &ctl->pblob;
 }
+
+uint64_t compel_get_leader_sp(struct parasite_ctl *ctl)
+{
+	return REG_SP(ctl->orig.regs);
+}
+
+uint64_t compel_get_thread_sp(struct parasite_thread_ctl *tctl)
+{
+	return REG_SP(tctl->th.regs);
+}
+
+uint64_t compel_get_leader_ip(struct parasite_ctl *ctl)
+{
+	return REG_IP(ctl->orig.regs);
+}
+
+uint64_t compel_get_thread_ip(struct parasite_thread_ctl *tctl)
+{
+	return REG_IP(tctl->th.regs);
+}
+
+void compel_set_leader_ip(struct parasite_ctl *ctl, uint64_t v)
+{
+	SET_REG_IP(ctl->orig.regs, v);
+}
+
+void compel_set_thread_ip(struct parasite_thread_ctl *tctl, uint64_t v)
+{
+	SET_REG_IP(tctl->th.regs, v);
+}
+
+void compel_get_stack(struct parasite_ctl *ctl, void **rstack, void **r_thread_stack)
+{
+	if (rstack)
+		*rstack = ctl->rstack;
+	if (r_thread_stack)
+		*r_thread_stack = ctl->r_thread_stack;
+}
