@@ -8,6 +8,7 @@
 
 #include "../soccr/soccr.h"
 
+#include "common/config.h"
 #include "cr_options.h"
 #include "util.h"
 #include "common/list.h"
@@ -19,7 +20,6 @@
 #include "image.h"
 #include "namespaces.h"
 #include "xmalloc.h"
-#include "config.h"
 #include "kerndat.h"
 #include "restorer.h"
 #include "rst-malloc.h"
@@ -27,11 +27,38 @@
 #include "protobuf.h"
 #include "images/tcp-stream.pb-c.h"
 
+#undef LOG_PREFIX
+#define LOG_PREFIX "tcp: "
 
 static LIST_HEAD(cpt_tcp_repair_sockets);
 static LIST_HEAD(rst_tcp_repair_sockets);
 
-static int tcp_repair_establised(int fd, struct inet_sk_desc *sk)
+static int lock_connection(struct inet_sk_desc *sk)
+{
+	if (opts.network_lock_method == NETWORK_LOCK_IPTABLES)
+		return iptables_lock_connection(sk);
+	else if (opts.network_lock_method == NETWORK_LOCK_NFTABLES)
+		return nftables_lock_connection(sk);
+	else if (opts.network_lock_method == NETWORK_LOCK_SKIP)
+		return 0;
+
+	return -1;
+}
+
+static int unlock_connection(struct inet_sk_desc *sk)
+{
+	if (opts.network_lock_method == NETWORK_LOCK_IPTABLES)
+		return iptables_unlock_connection(sk);
+	else if (opts.network_lock_method == NETWORK_LOCK_NFTABLES)
+		/* All connections will be unlocked in network_unlock(void) */
+		return 0;
+	else if (opts.network_lock_method == NETWORK_LOCK_SKIP)
+		return 0;
+
+	return -1;
+}
+
+static int tcp_repair_established(int fd, struct inet_sk_desc *sk)
 {
 	int ret;
 	struct libsoccr_sk *socr;
@@ -49,9 +76,11 @@ static int tcp_repair_establised(int fd, struct inet_sk_desc *sk)
 	}
 
 	if (!(root_ns_mask & CLONE_NEWNET)) {
-		ret = nf_lock_connection(sk);
-		if (ret < 0)
+		ret = lock_connection(sk);
+		if (ret < 0) {
+			pr_err("Failed to lock TCP connection %x\n", sk->sd.ino);
 			goto err2;
+		}
 	}
 
 	socr = libsoccr_pause(sk->rfd);
@@ -64,7 +93,7 @@ static int tcp_repair_establised(int fd, struct inet_sk_desc *sk)
 
 err3:
 	if (!(root_ns_mask & CLONE_NEWNET))
-		nf_unlock_connection(sk);
+		unlock_connection(sk);
 err2:
 	close(sk->rfd);
 err1:
@@ -78,9 +107,9 @@ static void tcp_unlock_one(struct inet_sk_desc *sk)
 	list_del(&sk->rlist);
 
 	if (!(root_ns_mask & CLONE_NEWNET)) {
-		ret = nf_unlock_connection(sk);
+		ret = unlock_connection(sk);
 		if (ret < 0)
-			pr_perror("Failed to unlock TCP connection");
+			pr_err("Failed to unlock TCP connection %x\n", sk->sd.ino);
 	}
 
 	libsoccr_resume(sk->priv);
@@ -118,8 +147,7 @@ static int dump_tcp_conn_state(struct inet_sk_desc *sk)
 		goto err_r;
 	}
 	if (ret != sizeof(data)) {
-		pr_err("This libsocr is not supported (%d vs %d)\n",
-				ret, (int)sizeof(data));
+		pr_err("This libsocr is not supported (%d vs %d)\n", ret, (int)sizeof(data));
 		goto err_r;
 	}
 
@@ -145,16 +173,16 @@ static int dump_tcp_conn_state(struct inet_sk_desc *sk)
 	}
 
 	if (data.flags & SOCCR_FLAGS_WINDOW) {
-		tse.has_snd_wl1		= true;
-		tse.has_snd_wnd		= true;
-		tse.has_max_window	= true;
-		tse.has_rcv_wnd		= true;
-		tse.has_rcv_wup		= true;
-		tse.snd_wl1		= data.snd_wl1;
-		tse.snd_wnd		= data.snd_wnd;
-		tse.max_window		= data.max_window;
-		tse.rcv_wnd		= data.rcv_wnd;
-		tse.rcv_wup		= data.rcv_wup;
+		tse.has_snd_wl1 = true;
+		tse.has_snd_wnd = true;
+		tse.has_max_window = true;
+		tse.has_rcv_wnd = true;
+		tse.has_rcv_wup = true;
+		tse.snd_wl1 = data.snd_wl1;
+		tse.snd_wnd = data.snd_wnd;
+		tse.max_window = data.max_window;
+		tse.rcv_wnd = data.rcv_wnd;
+		tse.rcv_wup = data.rcv_wup;
 	}
 
 	/*
@@ -216,14 +244,36 @@ err_r:
 	return ret;
 }
 
-int dump_one_tcp(int fd, struct inet_sk_desc *sk)
+int dump_one_tcp(int fd, struct inet_sk_desc *sk, SkOptsEntry *soe)
 {
+	soe->has_tcp_keepcnt = true;
+	if (dump_opt(fd, SOL_TCP, TCP_KEEPCNT, &soe->tcp_keepcnt)) {
+		pr_perror("Can't read TCP_KEEPCNT");
+		return -1;
+	}
+
+	soe->has_tcp_keepidle = true;
+	if (dump_opt(fd, SOL_TCP, TCP_KEEPIDLE, &soe->tcp_keepidle)) {
+		pr_perror("Can't read TCP_KEEPIDLE");
+		return -1;
+	}
+
+	soe->has_tcp_keepintvl = true;
+	if (dump_opt(fd, SOL_TCP, TCP_KEEPINTVL, &soe->tcp_keepintvl)) {
+		pr_perror("Can't read TCP_KEEPINTVL");
+		return -1;
+	}
+
 	if (sk->dst_port == 0)
 		return 0;
 
+	if (opts.tcp_close) {
+		return 0;
+	}
+
 	pr_info("Dumping TCP connection\n");
 
-	if (tcp_repair_establised(fd, sk))
+	if (tcp_repair_established(fd, sk))
 		return -1;
 
 	if (dump_tcp_conn_state(sk))
@@ -236,8 +286,7 @@ int dump_one_tcp(int fd, struct inet_sk_desc *sk)
 	return 0;
 }
 
-static int read_tcp_queue(struct libsoccr_sk *sk, struct libsoccr_sk_data *data,
-		int queue, u32 len, struct cr_img *img)
+static int read_tcp_queue(struct libsoccr_sk *sk, struct libsoccr_sk_data *data, int queue, u32 len, struct cr_img *img)
 {
 	char *buf;
 
@@ -292,7 +341,7 @@ static int restore_tcp_conn_state(int sk, struct libsoccr_sk *socr, struct inet_
 		goto err_c;
 	}
 
-	data.state = ii->ie->state;;
+	data.state = ii->ie->state;
 	data.inq_len = tse->inq_len;
 	data.inq_seq = tse->inq_seq;
 	data.outq_len = tse->outq_len;
@@ -327,13 +376,9 @@ static int restore_tcp_conn_state(int sk, struct libsoccr_sk *socr, struct inet_
 		data.rcv_wup = tse->rcv_wup;
 	}
 
-	if (restore_sockaddr(&sa_src,
-				ii->ie->family, ii->ie->src_port,
-				ii->ie->src_addr, 0) < 0)
+	if (restore_sockaddr(&sa_src, ii->ie->family, ii->ie->src_port, ii->ie->src_addr, 0) < 0)
 		goto err_c;
-	if (restore_sockaddr(&sa_dst,
-				ii->ie->family, ii->ie->dst_port,
-				ii->ie->dst_addr, 0) < 0)
+	if (restore_sockaddr(&sa_dst, ii->ie->family, ii->ie->dst_port, ii->ie->dst_addr, 0) < 0)
 		goto err_c;
 
 	libsoccr_set_addr(socr, 1, &sa_src, 0);
@@ -379,7 +424,7 @@ int prepare_tcp_socks(struct task_restore_args *ta)
 {
 	struct inet_sk_info *ii;
 
-	ta->tcp_socks = (struct rst_tcp_sock *) rst_mem_align_cpos(RM_PRIVATE);
+	ta->tcp_socks = (struct rst_tcp_sock *)rst_mem_align_cpos(RM_PRIVATE);
 	ta->tcp_socks_n = 0;
 
 	list_for_each_entry(ii, &rst_tcp_repair_sockets, rlist) {
@@ -387,7 +432,7 @@ int prepare_tcp_socks(struct task_restore_args *ta)
 
 		/*
 		 * rst_tcp_repair_sockets contains all sockets, so we need to
-		 * select sockets which restored in a current porcess.
+		 * select sockets which restored in a current process.
 		 */
 		if (ii->sk_fd == -1)
 			continue;
@@ -410,8 +455,10 @@ int restore_one_tcp(int fd, struct inet_sk_info *ii)
 
 	pr_info("Restoring TCP connection\n");
 
-	if (opts.tcp_close &&
-		ii->ie->state != TCP_LISTEN && ii->ie->state != TCP_CLOSE) {
+	if (opts.tcp_close) {
+		if (shutdown(fd, SHUT_RDWR) && errno != ENOTCONN) {
+			pr_perror("Unable to shutdown the socket id %x ino %x", ii->ie->id, ii->ie->ino);
+		}
 		return 0;
 	}
 
@@ -419,8 +466,10 @@ int restore_one_tcp(int fd, struct inet_sk_info *ii)
 	if (!sk)
 		return -1;
 
-	if (restore_tcp_conn_state(fd, sk, ii))
+	if (restore_tcp_conn_state(fd, sk, ii)) {
+		libsoccr_release(sk);
 		return -1;
+	}
 
 	return 0;
 }
@@ -431,14 +480,30 @@ void tcp_locked_conn_add(struct inet_sk_info *ii)
 	ii->sk_fd = -1;
 }
 
+static int unlock_connection_info(struct inet_sk_info *si)
+{
+	if (opts.network_lock_method == NETWORK_LOCK_IPTABLES)
+		return iptables_unlock_connection_info(si);
+	else if (opts.network_lock_method == NETWORK_LOCK_NFTABLES)
+		/* All connections will be unlocked in network_unlock(void) */
+		return 0;
+	else if (opts.network_lock_method == NETWORK_LOCK_SKIP)
+		return 0;
+
+	return -1;
+}
+
 void rst_unlock_tcp_connections(void)
 {
 	struct inet_sk_info *ii;
 
+	if (opts.tcp_close)
+		return;
+
 	/* Network will be unlocked by network-unlock scripts */
 	if (root_ns_mask & CLONE_NEWNET)
 		return;
 
 	list_for_each_entry(ii, &rst_tcp_repair_sockets, rlist)
-		nf_unlock_connection_info(ii);
+		unlock_connection_info(ii);
 }
